#ifndef FTPD_PRIVATE_H
#define FTPD_PRIVATE_H

#include <pkgconf/net_ftpserver.h>
#include <limits.h>
#include <sys/socket.h>

struct filename;
struct ftp_session;
struct reply;

typedef struct ftp_session {
	volatile int can_delete;

	ftp_server_t *server;
	struct ftp_session *next;

	int stack[CYGNUM_NET_FTPSERVER_STACK_SIZE / sizeof(int)];
	cyg_thread thread_obj;
	cyg_handle_t thread_hdl;

	unsigned int downloaded;	/* bytes downloaded */
	unsigned int uploaded;	/* bytes uploaded */

	int passive;
	int datafd;		/* data connection file descriptor */
	struct sockaddr_in ctrlconn;	/* stdin/stdout, for using the same ip number */


	char cmd[PATH_MAX + 32];	/* command line - about 30 chars for command */
	char wd[PATH_MAX + 1];	/* current working directory */
	int loggedin;
	char account[9];
	char * renamefrom;
	int epsvall; /* nonzero means we've seen an EPSV ALL */

	int debug;	/* don't give debug output */
	int guest;	/* if non-zero, it's a guest user */

	int ctrlsock; /* Control socket */

	int type;	/* type - 0 = error, 1 = ascii, 2 = binary */

	int restartat;
	unsigned int idletime;

	struct reply *lastreply;
	int replycode;

	struct sockaddr_in peer;
	unsigned short int peerdataport;

	union
	{
		char xferbuffer[CYGNUM_NET_FTPSERVER_XFER_BUFFER_SIZE]; /* Also used for path mangling */
		struct
		{
			char combpath[CYGNUM_NET_FTPSERVER_XFER_BUFFER_SIZE / 2];
			char fullpath[CYGNUM_NET_FTPSERVER_XFER_BUFFER_SIZE / 2];
		};
	};

	char cpybuffer[CYGNUM_NET_FTPSERVER_XFER_BUFFER_SIZE]; /* Also used for list output */

	int matches;
	int outptr;

	int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R, opt_r, opt_t, opt_S; /* ls options */

	int colwidth;
	int filenames;

	struct filename *head;
	struct filename *tail;
} ftp_session_t;

void donlist(ftp_session_t *, int, char * arg);

int opendata(ftp_session_t *);

void addreply( ftp_session_t *, int, const char *, ... );
void doreply( ftp_session_t * );

int combine_path(const char *old_dir, const char *path, char *result_path, int result_path_size);
int get_full_path(ftp_session_t *, const char *path, char *result_path, int result_path_size);
void ftpd_session_thrfunc( cyg_addrword_t );
int ftpd_session( ftp_session_t * );

#endif
